/**
 * Copyright (c) 2015-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 */
'use strict';

require('es6-map/implement');
require('es6-set/implement');

var test = require('tape');
var makeReporter = require('./reporter');
var spy = require('./spy');

var attachRenderer = require('../attachRenderer');
var globalHook = require('../GlobalHook.js');
globalHook(window);

if (!window.IS_TRAVIS) {
  makeReporter(test.createStream({ objectMode: true }));
}

var React = require('./v0.13/node_modules/react');

var _require = require('events');

var EventEmitter = _require.EventEmitter;

var renderers = window.__REACT_DEVTOOLS_GLOBAL_HOOK__._renderers;
var renderer = renderers[Object.keys(renderers)[0]];

function tracker(hook) {
  var els = new Map();
  var roots = new Set();
  hook.on('root', function (_ref) {
    var element = _ref.element;
    return roots.add(element);
  });
  hook.on('unmount', function (_ref2) {
    var element = _ref2.element;

    roots['delete'](element);
    els['delete'](element);
  });
  hook.on('mount', function (_ref3) {
    var element = _ref3.element;
    var data = _ref3.data;

    els.set(element, [data]);
  });
  hook.on('update', function (_ref4) {
    var element = _ref4.element;
    var data = _ref4.data;

    els.get(element).push(data);
  });
  return { els: els, roots: roots };
}

function setup(hook) {
  var handlers = {
    root: spy(),
    mount: spy(),
    update: spy(),
    unmount: spy() };
  for (var name in handlers) {
    hook.on(name, handlers[name]);
  }
  return handlers;
}

function wrapElement(hook, element) {
  var extras = attachRenderer(hook, 'abc', renderer);
  var node = document.createElement('div');
  React.render(element, node);
  extras.cleanup();
  React.unmountComponentAtNode(node);
}

function wrapRender(hook, fn) {
  var extras = attachRenderer(hook, 'abc', renderer);
  fn();
  extras.cleanup();
}

var SimpleApp = React.createClass({
  displayName: 'SimpleApp',

  render: function render() {
    return React.createElement(
      'div',
      null,
      'Hello'
    );
  } });

// Mounting and Unmounting

test('should work with plain DOM node', function (t) {
  var hook = new EventEmitter();
  var handlers = setup(hook);

  wrapElement(hook, React.createElement(
    'div',
    null,
    'Plain'
  ));

  t.ok(handlers.root.calledOnce, 'One root');
  // the root-level wrapper, and the div
  t.equal(handlers.mount.callCount, 2, 'Two mounts');
  t.notOk(handlers.unmount.called, 'No unmounts');
  t.end();
});

test('should work with simple composite component', function (t) {
  var hook = new EventEmitter();
  var handlers = setup(hook);

  wrapElement(hook, React.createElement(SimpleApp, null));

  t.ok(handlers.root.calledOnce, 'One root');
  // the root-level wrapper, the composite component, and the div
  t.equal(handlers.mount.callCount, 3, 'Three mounts');
  t.notOk(handlers.unmount.called, 'No unmounts');
  t.end();
});

test('attaching late should work', function (t) {
  var hook = new EventEmitter();
  var handlers = setup(hook);

  var node = document.createElement('div');
  React.render(React.createElement(SimpleApp, null), node);

  var extras = attachRenderer(hook, 'abc', renderer);
  extras.walkTree(function (component, data) {
    return handlers.mount({ component: component, data: data });
  }, function (component) {
    return handlers.root({ component: component });
  });

  t.equal(handlers.root.callCount, 1, 'One root');
  // the root-level wrapper, the composite component, and the div
  t.equal(handlers.mount.callCount, 3, 'Three mounts');
  t.notOk(handlers.unmount.called, 'No unmounts');

  // cleanup after
  extras.cleanup();
  React.unmountComponentAtNode(node);

  t.end();
});

test('should unmount everything', function (t) {
  var hook = new EventEmitter();
  var els = new Set();
  hook.on('mount', function (_ref5) {
    var element = _ref5.element;
    return els.add(element);
  });
  hook.on('unmount', function (_ref6) {
    var element = _ref6.element;
    return els['delete'](element);
  });

  wrapRender(hook, function () {
    var node = document.createElement('div');
    React.render(React.createElement(SimpleApp, null), node);
    t.ok(els.size > 0, 'Some elements');
    React.unmountComponentAtNode(node);
  });

  t.equal(els.size, 0, 'Everything unmounted');
  t.end();
});

test('should register two roots', function (t) {
  var hook = new EventEmitter();
  var handlers = setup(hook);

  wrapRender(hook, function () {
    var node = document.createElement('div');
    var node2 = document.createElement('div');
    React.render(React.createElement(SimpleApp, null), node);
    React.render(React.createElement(SimpleApp, null), node2);
    React.unmountComponentAtNode(node);
    React.unmountComponentAtNode(node2);
  });

  t.equal(handlers.root.callCount, 2, 'Two roots');
  t.end();
});

test('Double render', function (t) {
  var hook = new EventEmitter();
  var handlers = setup(hook);
  var els = new Set();
  hook.on('mount', function (_ref7) {
    var element = _ref7.element;
    return els.add(element);
  });
  hook.on('unmount', function (_ref8) {
    var element = _ref8.element;
    return els['delete'](element);
  });

  wrapNode(function (node) {
    wrapRender(hook, function () {
      React.render(React.createElement(SimpleApp, null), node);
      t.equal(handlers.update.callCount, 0, 'No updates');
      React.render(React.createElement(SimpleApp, null), node);
    });
  });

  t.equal(handlers.root.callCount, 1, 'One root');
  t.ok(handlers.update.callCount > 0, 'Updates');
  t.equal(els.size, 3, 'Only three mounted');
  t.end();
});

test('Plain text nodes', function (t) {
  var hook = new EventEmitter();

  var _tracker = tracker(hook);

  var roots = _tracker.roots;
  var els = _tracker.els;

  var PlainApp = React.createClass({
    displayName: 'PlainApp',

    render: function render() {
      return React.createElement(
        'div',
        null,
        'one',
        ['two'],
        'three'
      );
    } });
  wrapElement(hook, React.createElement(PlainApp, null));

  var root = roots.values().next().value;
  var composite = els.get(root)[0].children[0];
  var div = els.get(composite)[0].children[0];
  var texts = els.get(div)[0].children;

  var contents = ['one', 'two', 'three'];

  t.equals(texts.length, 3, '3 text children');

  texts.forEach(function (comp, i) {
    t.equals(els.get(comp)[0].text, contents[i], i + ') Text content correct');
    t.equals(els.get(comp)[0].nodeType, 'Text', i + ') NodeType = text');
  });

  t.end();
});

// State updating

var StateApp = React.createClass({
  displayName: 'StateApp',

  getInitialState: function getInitialState() {
    return { updated: false };
  },
  render: function render() {
    return React.createElement(
      'div',
      null,
      this.state.updated ? 'Updated' : 'Not updated'
    );
  } });

function wrapNode(fn) {
  var node = document.createElement('div');
  fn(node);
  React.unmountComponentAtNode(node);
}

test('State update', function (t) {
  var hook = new EventEmitter();

  var _tracker2 = tracker(hook);

  var roots = _tracker2.roots;
  var els = _tracker2.els;

  wrapNode(function (node) {
    wrapRender(hook, function () {
      var App = React.render(React.createElement(StateApp, null), node);
      App.setState({ updated: true });
    });
  });

  var composite = roots.values().next().value;
  var wrapper = els.get(composite)[0].children[0];
  var div = els.get(wrapper)[0].children[0];

  var divUpdates = els.get(div);
  t.equal(divUpdates[0].nodeType, 'Native', '[Div] Native type');
  t.equal(divUpdates[0].name, 'div', 'Named "div"');
  t.equal(divUpdates[0].children, 'Not updated', 'At first, not updated');
  t.equal(divUpdates[1].children, 'Updated', 'Then, updated');
  var updates = els.get(composite);
  t.equal(updates[0].nodeType, 'Composite', '[App] Composite type');
  t.equal(updates[0].name, 'StateApp', 'Named "StateApp"');
  t.equal(updates[0].state.updated, false, 'State[0] updated=false');
  t.equal(updates[1].state.updated, true, 'State[1] updated=true');
  t.end();
});

test('Props update', function (t) {
  var hook = new EventEmitter();

  var _tracker3 = tracker(hook);

  var roots = _tracker3.roots;
  var els = _tracker3.els;

  var StateProps = React.createClass({
    displayName: 'StateProps',

    getInitialState: function getInitialState() {
      return { pass: false };
    },
    render: function render() {
      return React.createElement(SimpleApp, { pass: this.state.pass });
    } });

  wrapNode(function (node) {
    wrapRender(hook, function () {
      var App = React.render(React.createElement(StateProps, null), node);
      App.setState({ pass: true });
      App.setState({ pass: 100 });
    });
  });

  var composite = roots.values().next().value;
  var simple = els.get(composite)[0].children[0];

  var updates = els.get(simple);
  t.equal(updates[0].props.pass, false, 't=0, prop=false');
  t.equal(updates[1].props.pass, true, 't=1, prop=true');
  t.equal(updates[2].props.pass, 100, 't=2, prop=100');
  t.end();
});